package io.theysay.affectr.client

import akka.actor.{Props, ActorSystem}
import spray.json.DefaultJsonProtocol
import spray.http.{BasicHttpCredentials, HttpRequest}
import concurrent.Future
import spray.io.IOExtension
import spray.can.client.HttpClient
import spray.client.HttpConduit

case class SentimentBias(positive: Option[Double], neutral: Option[Double], negative: Option[Double])
case class Request(text: String, level: Option[String] = None, ratio: Option[Double] = None,
                   targets: Option[String] = None, matching: Option[String] = None, bias: Option[SentimentBias] = None)

case class SentimentScore(label: String, positive: Double, negative: Double, neutral: Double, confidence: Option[Double]) {
  def getPolarity = label
}

case class SimpleSentiment(sentiment: SentimentScore, wordCount: Int)
case class SentenceSentiment(sentiment: SentimentScore, start: Int, end: Int, text: String)
case class WordPolarityTimeline(label: String, timelineY: Double)
case class WordSentiment(sentiment: WordPolarityTimeline, wordIndex: Int, text: String)

case class EntitySentiment(sentiment: SentimentScore, start: Int, end: Int, sentence: String, sentenceHtml: String, text: String, headNoun: String, headNounIndex: Int, salience: Int)

case class AggregatedEntity(sentiment: SentimentScore, start: Int, end: Int, sentence: String, sentenceHtml: String, text: String, headNoun: String, headNounIndex: Int, salience: Double)
case class AggregateEntitySentiment(category: String, frequency: Int, sentiment: SentimentScore, salience: Double, mentions: List[AggregatedEntity])

case class TaxonomyEntity(sentiment: SentimentScore, start: Int, end: Int, sentence: String, sentenceHtml: String, text: String, headNoun: String, headNounIndex: Int, salience: Double)
case class TaxonomyEntitySentiment(category: String, frequency: Int, sentiment: SentimentScore, salience: Double, mentions: List[TaxonomyEntity])

case class RelatedEntity(head: String, headIndex: Int, text: String)
case class EntityRelationSentiment(entity1: RelatedEntity, entity2: RelatedEntity, sentiment: SentimentScore, salience: Double, sentence: String, sentenceHtml: String)

case class Speculation(start: Int, end: Int, sentenceIndex: Int, speculationType: String, text: String)
case class Intent(start: Int, end: Int, sentenceIndex: Int, intentType: String, text: String)
case class Risk(start: Int, end: Int, sentenceIndex: Int, riskType: String, text: String)
case class Comparison(start: Int, end: Int, sentenceIndex: Int, comparisonType: String, text: String)

case class NamedEntity(head: String, headIndex: Int, start: Int, end: Int, sentence: String, sentenceHtml: String, text: String, namedEntityTypes: Array[String])
case class PosTag(posTag: String, posTaggedWord: String, sentenceIndex: Int, stem: String, text: String, wordIndex: Int)

case class Dependency(predicate: String, relation: String)
case class Dependent(text: String, stem: String, wordIndex: Int)
case class Governor(text: String, stem: String, wordIndex: Int)
case class DependencyParse(dependency: Dependency, dependent: Dependent, governor: Option[Governor])

case class Chunk(chunkType: String, start: Int, end: Int, text: String, sentenceIndex: Int)
case class ChunkHead(posTag: String, posTaggedWord: String, stem: String, text: String, wordIndex: Int)
case class ChunkConstituent(chunk: Chunk, head: ChunkHead)

import spray.httpx.SprayJsonSupport._

object AffectRJsonProtocol extends DefaultJsonProtocol {
  implicit val sentimentBiasFormat = jsonFormat3(SentimentBias)
  implicit val requestFormat = jsonFormat6(Request)

  implicit val sentimentScoreFormat = jsonFormat5(SentimentScore)
  implicit val simpleSentimentFormat = jsonFormat2(SimpleSentiment)
  implicit val sentenceSentimentFormat = jsonFormat4(SentenceSentiment)
  implicit val wordPolarityTimelineYFormat = jsonFormat2(WordPolarityTimeline)
  implicit val wordSentimentFormat = jsonFormat3(WordSentiment)

  implicit val entitySentimentFormat = jsonFormat9(EntitySentiment)

  implicit val aggregatedEntityFormat = jsonFormat9(AggregatedEntity)
  implicit val aggregatedEntitySentimentFormat = jsonFormat5(AggregateEntitySentiment)
  implicit val taxonomyEntityFormat = jsonFormat9(TaxonomyEntity)
  implicit val taxonomyEntitySentimentFormat = jsonFormat5(TaxonomyEntitySentiment)

  implicit val relatedEntityFormat = jsonFormat3(RelatedEntity)
  implicit val entityRelationSentimentFormat = jsonFormat6(EntityRelationSentiment)

  implicit val speculationFormat = jsonFormat5(Speculation)
  implicit val intentFormat = jsonFormat5(Intent)
  implicit val riskFormat = jsonFormat5(Risk)
  implicit val comparisonFormat = jsonFormat5(Comparison)

  implicit val namedEntityFormat = jsonFormat8(NamedEntity)
  implicit val posTagFormat = jsonFormat6(PosTag)

  implicit val dependencyFormat = jsonFormat2(Dependency)
  implicit val dependentFormat = jsonFormat3(Dependent)
  implicit val governorFormat = jsonFormat3(Governor)
  implicit val dependencyParseFormat = jsonFormat3(DependencyParse)

  implicit val chunkFormat = jsonFormat5(Chunk)
  implicit val chunkHeadFormat = jsonFormat5(ChunkHead)
  implicit val chunkConstituentFormat = jsonFormat2(ChunkConstituent)
}

class AffectR(actorSystem: ActorSystem, username: String, password: String) {
  implicit val ec = actorSystem.dispatcher

  val ioBridge = IOExtension(actorSystem).ioBridge()

  val httpClient = actorSystem.actorOf(
    props = Props(new HttpClient(ioBridge)),
    name = "http-client"
  )

  val conduit = actorSystem.actorOf(
    props = Props(new HttpConduit(httpClient, "api.theysay.io", 80)),
    name = "http-conduit"
  )

  import AffectRJsonProtocol._
  import HttpConduit._

  val simpleSentimentPipeline: HttpRequest => Future[SimpleSentiment] = (
    addCredentials(BasicHttpCredentials(username, password))
      ~> sendReceive(conduit)
      ~> unmarshal[SimpleSentiment]
    )

  val sentenceSentimentPipeline: HttpRequest => Future[Array[SentenceSentiment]] = (
    addCredentials(BasicHttpCredentials(username, password))
      ~> sendReceive(conduit)
      ~> unmarshal[Array[SentenceSentiment]]
    )

  val wordSentimentPipeline: HttpRequest => Future[Array[WordSentiment]] = (
    addCredentials(BasicHttpCredentials(username, password))
      ~> sendReceive(conduit)
      ~> unmarshal[Array[WordSentiment]]
    )

  val entitySentimentPipeline: HttpRequest => Future[Array[EntitySentiment]] = (
    addCredentials(BasicHttpCredentials(username, password))
      ~> sendReceive(conduit)
      ~> unmarshal[Array[EntitySentiment]]
    )

  val aggregatedEntitySentimentPipeline: HttpRequest => Future[Array[AggregateEntitySentiment]] = (
    addCredentials(BasicHttpCredentials(username, password))
      ~> sendReceive(conduit)
      ~> unmarshal[Array[AggregateEntitySentiment]]
    )

  val taxonomyEntitySentimentPipeline: HttpRequest => Future[Array[TaxonomyEntitySentiment]] = (
    addCredentials(BasicHttpCredentials(username, password))
      ~> sendReceive(conduit)
      ~> unmarshal[Array[TaxonomyEntitySentiment]]
    )

  val entityRelationSentimentPipeline: HttpRequest => Future[Array[EntityRelationSentiment]] = (
    addCredentials(BasicHttpCredentials(username, password))
      ~> sendReceive(conduit)
      ~> unmarshal[Array[EntityRelationSentiment]]
    )

  val speculationPipeline: HttpRequest => Future[Array[Speculation]] = (
    addCredentials(BasicHttpCredentials(username, password))
      ~> sendReceive(conduit)
      ~> unmarshal[Array[Speculation]]
    )

  val intentPipeline: HttpRequest => Future[Array[Intent]] = (
    addCredentials(BasicHttpCredentials(username, password))
      ~> sendReceive(conduit)
      ~> unmarshal[Array[Intent]]
    )

  val riskPipeline: HttpRequest => Future[Array[Risk]] = (
    addCredentials(BasicHttpCredentials(username, password))
      ~> sendReceive(conduit)
      ~> unmarshal[Array[Risk]]
    )

  val comparisonPipeline: HttpRequest => Future[Array[Comparison]] = (
    addCredentials(BasicHttpCredentials(username, password))
      ~> sendReceive(conduit)
      ~> unmarshal[Array[Comparison]]
    )

  val namedEntityPipeline: HttpRequest => Future[Array[NamedEntity]] = (
    addCredentials(BasicHttpCredentials(username, password))
      ~> sendReceive(conduit)
      ~> unmarshal[Array[NamedEntity]]
    )

  val posTagPipeline: HttpRequest => Future[Array[PosTag]] = (
    addCredentials(BasicHttpCredentials(username, password))
      ~> sendReceive(conduit)
      ~> unmarshal[Array[PosTag]]
    )

  val dependencyParsePipeline: HttpRequest => Future[Array[DependencyParse]] = (
    addCredentials(BasicHttpCredentials(username, password))
      ~> sendReceive(conduit)
      ~> unmarshal[Array[DependencyParse]]
    )

  val chunkParsePipeline: HttpRequest => Future[Array[ChunkConstituent]] = (
    addCredentials(BasicHttpCredentials(username, password))
      ~> sendReceive(conduit)
      ~> unmarshal[Array[ChunkConstituent]]
    )

  def classifySentiment(text: String, bias: Option[SentimentBias] = None): Future[SimpleSentiment] = {
    simpleSentimentPipeline(Post("/v1/sentiment", Request(text, bias = bias)))
  }

  def classifySentenceSentiment(text: String, bias: Option[SentimentBias] = None): Future[Array[SentenceSentiment]] = {
    sentenceSentimentPipeline(Post("/v1/sentiment", Request(text, Some("sentence"), bias = bias)))
  }

  def classifyWordSentiment(text: String, bias: Option[SentimentBias] = None): Future[Array[WordSentiment]] = {
    wordSentimentPipeline(Post("/v1/sentiment", Request(text, Some("word"), bias = bias)))
  }

  def classifyEntitySentiment(text: String, targets: Option[String] = None, matching: Option[String] = None, bias: Option[SentimentBias] = None): Future[Array[EntitySentiment]] = {
    entitySentimentPipeline(Post("/v1/sentiment", Request(text, Some("entity"), None, targets, matching, bias)))
  }

  def classifyAggregatedEntitySentiment(text: String, bias: Option[SentimentBias] = None): Future[Array[AggregateEntitySentiment]] = {
    aggregatedEntitySentimentPipeline(Post("/v1/sentiment", Request(text, Some("entityaggregate"), bias = bias)))
  }

  def classifyTaxonomyEntitySentiment(text: String, bias: Option[SentimentBias] = None): Future[Array[TaxonomyEntitySentiment]] = {
    taxonomyEntitySentimentPipeline(Post("/v1/sentiment", Request(text, Some("entitytaxonomy"), bias = bias)))
  }

  def classifyEntityRelationSentiment(text: String, bias: Option[SentimentBias] = None): Future[Array[EntityRelationSentiment]] = {
    entityRelationSentimentPipeline(Post("/v1/sentiment", Request(text, Some("entityrelation"), bias = bias)))
  }

  def classifySpeculation(text: String): Future[Array[Speculation]] = {
    speculationPipeline(Post("/v1/speculation", Request(text)))
  }

  def classifyIntent(text: String): Future[Array[Intent]] = {
    intentPipeline(Post("/v1/intent", Request(text)))
  }

  def classifyRisk(text: String): Future[Array[Risk]] = {
    riskPipeline(Post("/v1/risk", Request(text)))
  }

  def classifyComparison(text: String): Future[Array[Comparison]] = {
    comparisonPipeline(Post("/v1/comparison", Request(text)))
  }

  def getNamedEntities(text: String): Future[Array[NamedEntity]] = {
    namedEntityPipeline(Post("/v1/namedentity", Request(text)))
  }

  def posTag(text: String): Future[Array[PosTag]] = {
    posTagPipeline(Post("/v1/postag", Request(text)))
  }

  def dependencyParse(text: String): Future[Array[DependencyParse]] = {
    dependencyParsePipeline(Post("/v1/depparse", Request(text)))
  }

  def chunkParse(text: String): Future[Array[ChunkConstituent]] = {
    chunkParsePipeline(Post("/v1/chunkparse", Request(text)))
  }
}

object AffectR {
  def apply(actorSystem: ActorSystem, username: String, password: String) = new AffectR(actorSystem, username, password)
}